package repository

import (
	"context"
	"database/sql"
	"database/sql/driver"
	"errors"
	"git.sch9.ru/new_gate/ms-tester/internal/models"
	"git.sch9.ru/new_gate/ms-tester/pkg"
	"github.com/DATA-DOG/go-sqlmock"
	"github.com/jmoiron/sqlx"
	"github.com/stretchr/testify/require"
	"testing"
	"time"
)

// contestTestFixture encapsulates common test setup
type contestTestFixture struct {
	db          *sql.DB
	sqlxDB      *sqlx.DB
	mock        sqlmock.Sqlmock
	contestRepo *ContestRepository
}

// newContestTestFixture creates and initializes a new test fixture
func newContestTestFixture(t *testing.T) *contestTestFixture {
	db, mock, err := sqlmock.New(sqlmock.QueryMatcherOption(sqlmock.QueryMatcherEqual))
	require.NoError(t, err)

	sqlxDB := sqlx.NewDb(db, "sqlmock")
	contestRepo := NewContestRepository(sqlxDB)

	return &contestTestFixture{
		db:          db,
		sqlxDB:      sqlxDB,
		mock:        mock,
		contestRepo: contestRepo,
	}
}

// cleanup closes database connections
func (tf *contestTestFixture) cleanup() {
	tf.db.Close()
	tf.sqlxDB.Close()
}

func TestContestRepository_CreateContest(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid contest creation", func(t *testing.T) {
		title := "Contest title"
		rows := sqlmock.NewRows([]string{"id"}).AddRow(1)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(createContestQuery)).
			WithArgs(title).
			WillReturnRows(rows)

		id, err := tf.contestRepo.CreateContest(context.Background(), title)
		require.NoError(t, err)
		require.Equal(t, int32(1), id)
	})
}

func TestContestRepository_ReadContestById(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid contest read", func(t *testing.T) {
		contestID := int32(1)
		now := time.Now().UTC()
		expectedContest := &models.Contest{
			Id:        contestID,
			Title:     "Contest title",
			CreatedAt: now,
			UpdatedAt: now,
		}

		rows := sqlmock.NewRows([]string{"id", "title", "created_at", "updated_at"}).
			AddRow(expectedContest.Id, expectedContest.Title, expectedContest.CreatedAt, expectedContest.UpdatedAt)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readContestByIdQuery)).
			WithArgs(contestID).
			WillReturnRows(rows)

		contest, err := tf.contestRepo.ReadContestById(context.Background(), contestID)
		require.NoError(t, err)
		require.Equal(t, expectedContest, contest)
	})

	t.Run("contest not found", func(t *testing.T) {
		contestID := int32(999)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readContestByIdQuery)).
			WithArgs(contestID).
			WillReturnError(sql.ErrNoRows)

		contest, err := tf.contestRepo.ReadContestById(context.Background(), contestID)
		require.Error(t, err)
		require.ErrorIs(t, err, pkg.ErrNotFound)
		require.Nil(t, contest)
	})
}

func TestContestRepository_DeleteContest(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid contest deletion", func(t *testing.T) {
		id := int32(1)
		rows := sqlmock.NewResult(1, 1)

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(deleteContestQuery)).
			WithArgs(id).
			WillReturnResult(rows)

		err := tf.contestRepo.DeleteContest(context.Background(), id)
		require.NoError(t, err)
	})
}

func TestContestRepository_AddTask(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid task addition", func(t *testing.T) {
		taskID := int32(1)
		contestID := int32(1)
		rows := sqlmock.NewRows([]string{"id"}).AddRow(1)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(addTaskQuery)).
			WithArgs(taskID, contestID, contestID).
			WillReturnRows(rows)

		id, err := tf.contestRepo.AddTask(context.Background(), contestID, taskID)
		require.NoError(t, err)
		require.Equal(t, int32(1), id)
	})
}

func TestContestRepository_DeleteTask(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid task deletion", func(t *testing.T) {
		id := int32(1)
		rows := sqlmock.NewResult(1, 1)

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(deleteTaskQuery)).
			WithArgs(id).
			WillReturnResult(rows)

		err := tf.contestRepo.DeleteTask(context.Background(), id)
		require.NoError(t, err)
	})
}

func TestContestRepository_AddParticipant(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid participant addition", func(t *testing.T) {
		contestID := int32(1)
		userID := int32(1)
		rows := sqlmock.NewRows([]string{"id"}).AddRow(1)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(addParticipantQuery)).
			WithArgs(contestID, userID, "").
			WillReturnRows(rows)

		id, err := tf.contestRepo.AddParticipant(context.Background(), contestID, userID)
		require.NoError(t, err)
		require.Equal(t, int32(1), id)
	})
}

func TestContestRepository_DeleteParticipant(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid participant deletion", func(t *testing.T) {
		id := int32(1)
		rows := sqlmock.NewResult(1, 1)

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(deleteParticipantQuery)).
			WithArgs(id).
			WillReturnResult(rows)

		err := tf.contestRepo.DeleteParticipant(context.Background(), id)
		require.NoError(t, err)
	})
}

func TestContestRepository_ReadTasks(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("successful tasks retrieval", func(t *testing.T) {
		contestID := int32(1)
		now := time.Now().UTC()
		expectedTasks := []*models.TasksListItem{
			{
				Id:          1,
				ProblemId:   10,
				ContestId:   contestID,
				Position:    1,
				Title:       "Task 1",
				MemoryLimit: 256,
				TimeLimit:   1000,
				CreatedAt:   now,
				UpdatedAt:   now,
			},
			{
				Id:          2,
				ProblemId:   11,
				ContestId:   contestID,
				Position:    2,
				Title:       "Task 2",
				MemoryLimit: 512,
				TimeLimit:   2000,
				CreatedAt:   now,
				UpdatedAt:   now,
			},
		}

		rows := sqlmock.NewRows([]string{
			"id", "problem_id", "contest_id", "position", "title",
			"memory_limit", "time_limit", "created_at", "updated_at",
		}).
			AddRow(
				expectedTasks[0].Id, expectedTasks[0].ProblemId, expectedTasks[0].ContestId,
				expectedTasks[0].Position, expectedTasks[0].Title, expectedTasks[0].MemoryLimit,
				expectedTasks[0].TimeLimit, expectedTasks[0].CreatedAt, expectedTasks[0].UpdatedAt,
			).
			AddRow(
				expectedTasks[1].Id, expectedTasks[1].ProblemId, expectedTasks[1].ContestId,
				expectedTasks[1].Position, expectedTasks[1].Title, expectedTasks[1].MemoryLimit,
				expectedTasks[1].TimeLimit, expectedTasks[1].CreatedAt, expectedTasks[1].UpdatedAt,
			)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readTasksQuery)).
			WithArgs(contestID).
			WillReturnRows(rows)

		tasks, err := tf.contestRepo.ReadTasks(context.Background(), contestID)
		require.NoError(t, err)
		require.Equal(t, expectedTasks, tasks)
	})

	t.Run("no tasks found", func(t *testing.T) {
		contestID := int32(999)

		rows := sqlmock.NewRows([]string{
			"id", "problem_id", "contest_id", "position", "title",
			"memory_limit", "time_limit", "created_at", "updated_at",
		})

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readTasksQuery)).
			WithArgs(contestID).
			WillReturnRows(rows)

		tasks, err := tf.contestRepo.ReadTasks(context.Background(), contestID)
		require.NoError(t, err)
		require.Empty(t, tasks)
	})

	t.Run("database error", func(t *testing.T) {
		contestID := int32(1)
		expectedErr := sql.ErrConnDone

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readTasksQuery)).
			WithArgs(contestID).
			WillReturnError(expectedErr)

		tasks, err := tf.contestRepo.ReadTasks(context.Background(), contestID)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ReadTasks")
		require.Nil(t, tasks)
	})
}

func TestContestRepository_ListContests(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("successful contests list retrieval", func(t *testing.T) {
		filter := models.ContestsFilter{
			Page:     1,
			PageSize: 2,
		}
		now := time.Now().UTC()
		expectedContests := []*models.ContestsListItem{
			{
				Id:        1,
				Title:     "Contest 1",
				CreatedAt: now,
				UpdatedAt: now,
			},
			{
				Id:        2,
				Title:     "Contest 2",
				CreatedAt: now,
				UpdatedAt: now,
			},
		}
		totalCount := int32(5)

		// Mock contests query
		contestsRows := sqlmock.NewRows([]string{"id", "title", "created_at", "updated_at"}).
			AddRow(
				expectedContests[0].Id, expectedContests[0].Title,
				expectedContests[0].CreatedAt, expectedContests[0].UpdatedAt,
			).
			AddRow(
				expectedContests[1].Id, expectedContests[1].Title,
				expectedContests[1].CreatedAt, expectedContests[1].UpdatedAt,
			)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readContestsListQuery)).
			WithArgs(filter.PageSize, filter.Offset()).
			WillReturnRows(contestsRows)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(totalCount)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countContestsQuery)).
			WillReturnRows(countRows)

		result, err := tf.contestRepo.ListContests(context.Background(), filter)
		require.NoError(t, err)
		expectedResult := &models.ContestsList{
			Contests: expectedContests,
			Pagination: models.Pagination{
				Total: models.Total(totalCount, filter.PageSize),
				Page:  filter.Page,
			},
		}
		require.Equal(t, expectedResult, result)
	})

	t.Run("no contests found", func(t *testing.T) {
		filter := models.ContestsFilter{
			Page:     1,
			PageSize: 2,
		}

		// Mock empty contests query
		contestsRows := sqlmock.NewRows([]string{"id", "title", "created_at", "updated_at"})

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readContestsListQuery)).
			WithArgs(filter.PageSize, filter.Offset()).
			WillReturnRows(contestsRows)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(int32(0))

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countContestsQuery)).
			WillReturnRows(countRows)

		result, err := tf.contestRepo.ListContests(context.Background(), filter)
		require.NoError(t, err)
		expectedResult := &models.ContestsList{
			Contests: []*models.ContestsListItem{},
			Pagination: models.Pagination{
				Total: models.Total(0, filter.PageSize),
				Page:  filter.Page,
			},
		}
		require.Equal(t, expectedResult, result)
	})

	t.Run("error in contests query", func(t *testing.T) {
		filter := models.ContestsFilter{
			Page:     1,
			PageSize: 2,
		}
		expectedErr := sql.ErrConnDone

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readContestsListQuery)).
			WithArgs(filter.PageSize, filter.Offset()).
			WillReturnError(expectedErr)

		result, err := tf.contestRepo.ListContests(context.Background(), filter)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ReadTasks")
		require.Nil(t, result)
	})

	t.Run("error in count query", func(t *testing.T) {
		filter := models.ContestsFilter{
			Page:     1,
			PageSize: 2,
		}
		expectedErr := sql.ErrConnDone

		// Mock successful contests query
		contestsRows := sqlmock.NewRows([]string{"id", "title", "created_at", "updated_at"})

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readContestsListQuery)).
			WithArgs(filter.PageSize, filter.Offset()).
			WillReturnRows(contestsRows)

		// Mock failing count query
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countContestsQuery)).
			WillReturnError(expectedErr)

		result, err := tf.contestRepo.ListContests(context.Background(), filter)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ReadTasks")
		require.Nil(t, result)
	})
}

func TestContestRepository_ListContestParticipants(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("successful participants list retrieval", func(t *testing.T) {
		filter := models.ParticipantsFilter{
			ContestId: 1,
			Page:      1,
			PageSize:  2,
		}
		now := time.Now().UTC()
		expectedParticipants := []*models.ParticipantsListItem{
			{
				Id:        1,
				UserId:    101,
				Name:      "Participant 1",
				CreatedAt: now,
				UpdatedAt: now,
			},
			{
				Id:        2,
				UserId:    102,
				Name:      "Participant 2",
				CreatedAt: now,
				UpdatedAt: now,
			},
		}
		totalCount := int32(5)

		// Mock participants query
		participantsRows := sqlmock.NewRows([]string{"id", "user_id", "name", "created_at", "updated_at"}).
			AddRow(
				expectedParticipants[0].Id, expectedParticipants[0].UserId, expectedParticipants[0].Name,
				expectedParticipants[0].CreatedAt, expectedParticipants[0].UpdatedAt,
			).
			AddRow(
				expectedParticipants[1].Id, expectedParticipants[1].UserId, expectedParticipants[1].Name,
				expectedParticipants[1].CreatedAt, expectedParticipants[1].UpdatedAt,
			)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readParticipantsListQuery)).
			WithArgs(filter.ContestId, filter.PageSize, filter.Offset()).
			WillReturnRows(participantsRows)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(totalCount)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countParticipantsQuery)).
			WithArgs(filter.ContestId).
			WillReturnRows(countRows)

		result, err := tf.contestRepo.ListParticipants(context.Background(), filter)
		require.NoError(t, err)
		expectedResult := &models.ParticipantsList{
			Participants: expectedParticipants,
			Pagination: models.Pagination{
				Total: models.Total(totalCount, filter.PageSize),
				Page:  filter.Page,
			},
		}
		require.Equal(t, expectedResult, result)
	})

	t.Run("no participants found", func(t *testing.T) {
		filter := models.ParticipantsFilter{
			ContestId: 999,
			Page:      1,
			PageSize:  2,
		}

		// Mock empty participants query
		participantsRows := sqlmock.NewRows([]string{"id", "user_id", "name", "created_at", "updated_at"})

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readParticipantsListQuery)).
			WithArgs(filter.ContestId, filter.PageSize, filter.Offset()).
			WillReturnRows(participantsRows)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(int32(0))

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countParticipantsQuery)).
			WithArgs(filter.ContestId).
			WillReturnRows(countRows)

		result, err := tf.contestRepo.ListParticipants(context.Background(), filter)
		require.NoError(t, err)
		expectedResult := &models.ParticipantsList{
			Participants: []*models.ParticipantsListItem{},
			Pagination: models.Pagination{
				Total: models.Total(0, filter.PageSize),
				Page:  filter.Page,
			},
		}
		require.Equal(t, expectedResult, result)
	})

	t.Run("page size limited to 1 when exceeding 20", func(t *testing.T) {
		filter := models.ParticipantsFilter{
			ContestId: 1,
			Page:      1,
			PageSize:  25, // Exceeds limit
		}
		now := time.Now().UTC()
		expectedParticipants := []*models.ParticipantsListItem{
			{
				Id:        1,
				UserId:    101,
				Name:      "Participant 1",
				CreatedAt: now,
				UpdatedAt: now,
			},
		}
		totalCount := int32(5)

		// Mock participants query with limited page size (1)
		participantsRows := sqlmock.NewRows([]string{"id", "user_id", "name", "created_at", "updated_at"}).
			AddRow(
				expectedParticipants[0].Id, expectedParticipants[0].UserId, expectedParticipants[0].Name,
				expectedParticipants[0].CreatedAt, expectedParticipants[0].UpdatedAt,
			)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readParticipantsListQuery)).
			WithArgs(filter.ContestId, int32(1), filter.Offset()). // PageSize limited to 1
			WillReturnRows(participantsRows)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(totalCount)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countParticipantsQuery)).
			WithArgs(filter.ContestId).
			WillReturnRows(countRows)

		result, err := tf.contestRepo.ListParticipants(context.Background(), filter)
		require.NoError(t, err)
		expectedResult := &models.ParticipantsList{
			Participants: expectedParticipants,
			Pagination: models.Pagination{
				Total: models.Total(totalCount, 1), // PageSize limited to 1
				Page:  filter.Page,
			},
		}
		require.Equal(t, expectedResult, result)
	})

	t.Run("error in participants query", func(t *testing.T) {
		filter := models.ParticipantsFilter{
			ContestId: 1,
			Page:      1,
			PageSize:  2,
		}
		expectedErr := sql.ErrConnDone

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readParticipantsListQuery)).
			WithArgs(filter.ContestId, filter.PageSize, filter.Offset()).
			WillReturnError(expectedErr)

		result, err := tf.contestRepo.ListParticipants(context.Background(), filter)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ReadParticipants")
		require.Nil(t, result)
	})

	t.Run("error in count query", func(t *testing.T) {
		filter := models.ParticipantsFilter{
			ContestId: 1,
			Page:      1,
			PageSize:  2,
		}
		expectedErr := sql.ErrConnDone

		// Mock successful participants query
		participantsRows := sqlmock.NewRows([]string{"id", "user_id", "name", "created_at", "updated_at"})

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readParticipantsListQuery)).
			WithArgs(filter.ContestId, filter.PageSize, filter.Offset()).
			WillReturnRows(participantsRows)

		// Mock failing count query
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countParticipantsQuery)).
			WithArgs(filter.ContestId).
			WillReturnError(expectedErr)

		result, err := tf.contestRepo.ListParticipants(context.Background(), filter)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ReadParticipants")
		require.Nil(t, result)
	})
}

func TestContestRepository_UpdateContest(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("successful contest update", func(t *testing.T) {
		contestID := int32(1)
		update := models.ContestUpdate{
			Title: sp("Updated Contest"),
		}
		result := sqlmock.NewResult(1, 1)

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(updateContestQuery)).
			WithArgs(update.Title, contestID).
			WillReturnResult(result)

		err := tf.contestRepo.UpdateContest(context.Background(), contestID, update)
		require.NoError(t, err)
	})

	t.Run("database error", func(t *testing.T) {
		contestID := int32(1)
		update := models.ContestUpdate{
			Title: sp("Updated Contest"),
		}
		expectedErr := sql.ErrConnDone

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(updateContestQuery)).
			WithArgs(update.Title, contestID).
			WillReturnError(expectedErr)

		err := tf.contestRepo.UpdateContest(context.Background(), contestID, update)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.UpdateContest")
	})
}

func TestContestRepository_UpdateParticipant(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("successful participant update", func(t *testing.T) {
		participantID := int32(1)
		update := models.ParticipantUpdate{
			Name: sp("Updated Name"),
		}
		result := sqlmock.NewResult(1, 1)

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(updateParticipantQuery)).
			WithArgs(update.Name, participantID).
			WillReturnResult(result)

		err := tf.contestRepo.UpdateParticipant(context.Background(), participantID, update)
		require.NoError(t, err)
	})

	t.Run("database error", func(t *testing.T) {
		participantID := int32(1)
		update := models.ParticipantUpdate{
			Name: sp("Updated Name"),
		}
		expectedErr := sql.ErrConnDone

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(updateParticipantQuery)).
			WithArgs(update.Name, participantID).
			WillReturnError(expectedErr)

		err := tf.contestRepo.UpdateParticipant(context.Background(), participantID, update)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.UpdateParticipant")
	})
}

func TestContestRepository_ReadSolution(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("successful solution read", func(t *testing.T) {
		solutionID := int32(1)
		now := time.Now().UTC()
		expectedSolution := &models.Solution{
			Id:            solutionID,
			TaskId:        10,
			ParticipantId: 20,
			Language:      228, // FIXME: use constant
			Penalty:       0,
			Solution:      "func main() {}",
			CreatedAt:     now,
			UpdatedAt:     now,
		}

		rows := sqlmock.NewRows([]string{
			"id", "task_id", "participant_id", "language", "penalty", "solution", "created_at", "updated_at",
		}).
			AddRow(
				expectedSolution.Id, expectedSolution.TaskId, expectedSolution.ParticipantId,
				expectedSolution.Language, expectedSolution.Penalty, expectedSolution.Solution,
				expectedSolution.CreatedAt, expectedSolution.UpdatedAt,
			)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readSolutionQuery)).
			WithArgs(solutionID).
			WillReturnRows(rows)

		solution, err := tf.contestRepo.ReadSolution(context.Background(), solutionID)
		require.NoError(t, err)
		require.Equal(t, expectedSolution, solution)
	})

	t.Run("solution not found", func(t *testing.T) {
		solutionID := int32(999)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readSolutionQuery)).
			WithArgs(solutionID).
			WillReturnError(sql.ErrNoRows)

		solution, err := tf.contestRepo.ReadSolution(context.Background(), solutionID)
		require.Error(t, err)
		require.ErrorIs(t, err, pkg.ErrNotFound)
		require.Nil(t, solution)
	})

	t.Run("database error", func(t *testing.T) {
		solutionID := int32(1)
		expectedErr := sql.ErrConnDone

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readSolutionQuery)).
			WithArgs(solutionID).
			WillReturnError(expectedErr)

		solution, err := tf.contestRepo.ReadSolution(context.Background(), solutionID)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ReadSolution")
		require.Nil(t, solution)
	})
}

func TestContestRepository_CreateSolution(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("successful solution creation", func(t *testing.T) {
		creation := &models.SolutionCreation{
			TaskId:        10,
			ParticipantId: 20,
			Language:      228, // FIXME: use constant
			Penalty:       0,
			Solution:      "func main() {}",
		}
		expectedID := int32(1)

		rows := sqlmock.NewRows([]string{"id"}).AddRow(expectedID)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(createSolutionQuery)).
			WithArgs(creation.TaskId, creation.ParticipantId, creation.Language, creation.Penalty, creation.Solution).
			WillReturnRows(rows)

		id, err := tf.contestRepo.CreateSolution(context.Background(), creation)
		require.NoError(t, err)
		require.Equal(t, expectedID, id)
	})

	t.Run("database error on query", func(t *testing.T) {
		creation := &models.SolutionCreation{
			TaskId:        10,
			ParticipantId: 20,
			Language:      228, // FIXME: use constant
			Penalty:       0,
			Solution:      "func main() {}",
		}
		expectedErr := sql.ErrConnDone

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(createSolutionQuery)).
			WithArgs(creation.TaskId, creation.ParticipantId, creation.Language, creation.Penalty, creation.Solution).
			WillReturnError(expectedErr)

		id, err := tf.contestRepo.CreateSolution(context.Background(), creation)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.CreateSolution")
		require.Equal(t, int32(0), id)
	})

	t.Run("database error on scan", func(t *testing.T) {
		creation := &models.SolutionCreation{
			TaskId:        10,
			ParticipantId: 20,
			Language:      228, // FIXME: use constant
			Penalty:       0,
			Solution:      "func main() {}",
		}
		rows := sqlmock.NewRows([]string{"id"}).AddRow(nil) // Invalid row to cause scan error

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(createSolutionQuery)).
			WithArgs(creation.TaskId, creation.ParticipantId, creation.Language, creation.Penalty, creation.Solution).
			WillReturnRows(rows)

		id, err := tf.contestRepo.CreateSolution(context.Background(), creation)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.CreateSolution")
		require.Equal(t, int32(0), id)
	})
}

func TestContestRepository_ListSolutions(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	// Define columns once for all test cases
	solutionColumns := []string{
		"id", "participant_id", "participant_name", "state", "score", "penalty",
		"time_stat", "memory_stat", "language", "task_id", "task_position",
		"task_title", "contest_id", "contest_title", "updated_at", "created_at",
	}

	t.Run("successful solutions list retrieval with filters", func(t *testing.T) {
		filter := models.SolutionsFilter{
			ContestId:     i32p(int32(1)),
			ParticipantId: i32p(int32(101)),
			TaskId:        i32p(int32(10)),
			Language:      i32p(int32(228)),
			State:         i32p(int32(1)), // Assuming state is an int32
			Page:          1,
			PageSize:      2,
			Order:         i32p(int32(-1)), // DESC order
		}
		now := time.Now().UTC()
		expectedSolutions := []*models.SolutionsListItem{
			{
				Id:              1,
				ParticipantId:   101,
				ParticipantName: "Participant 1",
				State:           1,
				Score:           100,
				Penalty:         0,
				TimeStat:        500,
				MemoryStat:      1024,
				Language:        228,
				TaskId:          10,
				TaskPosition:    1,
				TaskTitle:       "Task 1",
				ContestId:       1,
				ContestTitle:    "Contest 1",
				UpdatedAt:       now,
				CreatedAt:       now,
			},
			{
				Id:              2,
				ParticipantId:   101,
				ParticipantName: "Participant 1",
				State:           1,
				Score:           90,
				Penalty:         0,
				TimeStat:        600,
				MemoryStat:      2048,
				Language:        228,
				TaskId:          10,
				TaskPosition:    1,
				TaskTitle:       "Task 1",
				ContestId:       1,
				ContestTitle:    "Contest 1",
				UpdatedAt:       now,
				CreatedAt:       now,
			},
		}
		totalCount := int32(5)

		// Build queries using the function
		baseQb, countQb := buildListSolutionsQueries(filter)
		countQuery, countArgs, err := countQb.ToSql()
		require.NoError(t, err)
		baseQuery, baseArgs, err := baseQb.ToSql()
		require.NoError(t, err)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(totalCount)
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countQuery)).
			WithArgs(i2v(countArgs)...).
			WillReturnRows(countRows)

		// Mock solutions query
		solutionsRows := sqlmock.NewRows(solutionColumns)
		for _, sol := range expectedSolutions {
			solutionsRows.AddRow(
				sol.Id, sol.ParticipantId, sol.ParticipantName, sol.State, sol.Score, sol.Penalty,
				sol.TimeStat, sol.MemoryStat, sol.Language, sol.TaskId, sol.TaskPosition,
				sol.TaskTitle, sol.ContestId, sol.ContestTitle, sol.UpdatedAt, sol.CreatedAt,
			)
		}

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(baseQuery)).
			WithArgs(i2v(baseArgs)...).
			WillReturnRows(solutionsRows)

		result, err := tf.contestRepo.ListSolutions(context.Background(), filter)
		require.NoError(t, err)
		expectedResult := &models.SolutionsList{
			Solutions: expectedSolutions,
			Pagination: models.Pagination{
				Total: models.Total(totalCount, filter.PageSize),
				Page:  filter.Page,
			},
		}
		require.Equal(t, expectedResult, result)
	})

	t.Run("no solutions found", func(t *testing.T) {
		filter := models.SolutionsFilter{
			ContestId: i32p(int32(999)),
			Page:      1,
			PageSize:  2,
		}

		// Build queries using the function
		baseQb, countQb := buildListSolutionsQueries(filter)
		countQuery, countArgs, err := countQb.ToSql()
		require.NoError(t, err)
		baseQuery, baseArgs, err := baseQb.ToSql()
		require.NoError(t, err)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(int32(0))
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countQuery)).
			WithArgs(i2v(countArgs)...).
			WillReturnRows(countRows)

		// Mock solutions query
		solutionsRows := sqlmock.NewRows(solutionColumns)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(baseQuery)).
			WithArgs(i2v(baseArgs)...).
			WillReturnRows(solutionsRows)

		result, err := tf.contestRepo.ListSolutions(context.Background(), filter)
		require.NoError(t, err)
		expectedResult := &models.SolutionsList{
			Solutions: []*models.SolutionsListItem{},
			Pagination: models.Pagination{
				Total: models.Total(0, filter.PageSize),
				Page:  filter.Page,
			},
		}
		require.Equal(t, expectedResult, result)
	})

	t.Run("error in count query", func(t *testing.T) {
		filter := models.SolutionsFilter{
			ContestId: i32p(int32(1)),
			Page:      1,
			PageSize:  2,
		}
		expectedErr := sql.ErrConnDone

		// Build queries using the function
		_, countQb := buildListSolutionsQueries(filter)
		countQuery, countArgs, err := countQb.ToSql()
		require.NoError(t, err)

		// Mock count query with error
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countQuery)).
			WithArgs(i2v(countArgs)...).
			WillReturnError(expectedErr)

		result, err := tf.contestRepo.ListSolutions(context.Background(), filter)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ListSolutions")
		require.Nil(t, result)
	})

	t.Run("error in solutions query", func(t *testing.T) {
		filter := models.SolutionsFilter{
			ContestId: i32p(int32(1)),
			Page:      1,
			PageSize:  2,
		}
		totalCount := int32(5)
		expectedErr := sql.ErrConnDone

		// Build queries using the function
		baseQb, countQb := buildListSolutionsQueries(filter)
		countQuery, countArgs, err := countQb.ToSql()
		require.NoError(t, err)
		baseQuery, baseArgs, err := baseQb.ToSql()
		require.NoError(t, err)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(totalCount)
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countQuery)).
			WithArgs(i2v(countArgs)...).
			WillReturnRows(countRows)

		// Mock solutions query with error
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(baseQuery)).
			WithArgs(i2v(baseArgs)...).
			WillReturnError(expectedErr)

		result, err := tf.contestRepo.ListSolutions(context.Background(), filter)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ListSolutions")
		require.Nil(t, result)
	})

	t.Run("error in struct scan", func(t *testing.T) {
		filter := models.SolutionsFilter{
			ContestId: i32p(int32(1)),
			Page:      1,
			PageSize:  2,
		}
		totalCount := int32(1)

		// Build queries using the function
		baseQb, countQb := buildListSolutionsQueries(filter)
		countQuery, countArgs, err := countQb.ToSql()
		require.NoError(t, err)
		baseQuery, baseArgs, err := baseQb.ToSql()
		require.NoError(t, err)

		// Mock count query
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(totalCount)
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(countQuery)).
			WithArgs(i2v(countArgs)...).
			WillReturnRows(countRows)

		// Mock solutions query with invalid data to cause scan error
		solutionsRows := sqlmock.NewRows(solutionColumns).AddRow(
			nil, 101, "Participant 1", 1, 100, 0, // Invalid id (nil)
			500, 1024, 228, 10, 1, "Task 1", 1, "Contest 1",
			time.Now().UTC(), time.Now().UTC(),
		)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(baseQuery)).
			WithArgs(i2v(baseArgs)...).
			WillReturnRows(solutionsRows)

		result, err := tf.contestRepo.ListSolutions(context.Background(), filter)
		require.Error(t, err)
		require.Contains(t, err.Error(), "ContestRepository.ListSolutions")
		require.Nil(t, result)
	})
}

func TestContestRepository_ReadTask(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid task read", func(t *testing.T) {
		taskID := int32(1)
		now := time.Now().UTC()
		expectedTask := &models.Task{
			Id:               taskID,
			Position:         1,
			Title:            "Task title",
			TimeLimit:        1000,
			MemoryLimit:      256,
			ProblemId:        1,
			ContestId:        1,
			LegendHtml:       "<p>Legend</p>",
			InputFormatHtml:  "<p>Input</p>",
			OutputFormatHtml: "<p>Output</p>",
			NotesHtml:        "<p>Notes</p>",
			ScoringHtml:      "<p>Scoring</p>",
			CreatedAt:        now,
			UpdatedAt:        now,
		}

		rows := sqlmock.NewRows([]string{
			"id",
			"position",
			"title",
			"time_limit",
			"memory_limit",
			"problem_id",
			"contest_id",
			"legend_html",
			"input_format_html",
			"output_format_html",
			"notes_html",
			"scoring_html",
			"created_at",
			"updated_at",
		}).AddRow(
			expectedTask.Id,
			expectedTask.Position,
			expectedTask.Title,
			expectedTask.TimeLimit,
			expectedTask.MemoryLimit,
			expectedTask.ProblemId,
			expectedTask.ContestId,
			expectedTask.LegendHtml,
			expectedTask.InputFormatHtml,
			expectedTask.OutputFormatHtml,
			expectedTask.NotesHtml,
			expectedTask.ScoringHtml,
			expectedTask.CreatedAt,
			expectedTask.UpdatedAt,
		)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readTaskQuery)).
			WithArgs(taskID).
			WillReturnRows(rows)

		task, err := tf.contestRepo.ReadTask(context.Background(), taskID)
		require.NoError(t, err)
		require.Equal(t, expectedTask, task)
	})

	t.Run("task not found", func(t *testing.T) {
		taskID := int32(999)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readTaskQuery)).
			WithArgs(taskID).
			WillReturnError(sql.ErrNoRows)

		task, err := tf.contestRepo.ReadTask(context.Background(), taskID)
		require.Error(t, err)
		require.ErrorIs(t, err, pkg.ErrNotFound)
		require.Nil(t, task)
	})
}

func TestContestRepository_ReadMonitor(t *testing.T) {
	tf := newContestTestFixture(t)
	defer tf.cleanup()

	t.Run("valid monitor read", func(t *testing.T) {
		contestID := int32(1)
		now := time.Now().UTC()
		penalty := int32(20)

		// Expected monitor data
		expectedMonitor := &models.Monitor{
			Summary: []*models.ProblemStatSummary{
				{
					Id:       1,
					Position: 1,
					Total:    10,
					Success:  5,
				},
				{
					Id:       2,
					Position: 2,
					Total:    8,
					Success:  3,
				},
			},
			Participants: []*models.ParticipantsStat{
				{
					Id:             1,
					Name:           "Participant 1",
					SolvedInTotal:  2,
					PenaltyInTotal: 40,
					Solutions: []*models.SolutionsListItem{
						{
							Id:              101,
							ParticipantId:   1,
							ParticipantName: "Participant 1",
							State:           5,
							Score:           100,
							Penalty:         0,
							TimeStat:        1000,
							MemoryStat:      256,
							Language:        228, // FIXME: use constants
							TaskId:          1,
							TaskPosition:    1,
							TaskTitle:       "Task 1",
							ContestId:       1,
							ContestTitle:    "Contest 1",
							CreatedAt:       now,
							UpdatedAt:       now,
						},
					},
				},
			},
		}

		// Mock statistics query
		statsRows := sqlmock.NewRows([]string{"task_id", "position", "total", "success"}).
			AddRow(1, 1, 10, 5).
			AddRow(2, 2, 8, 3)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readStatisticsQuery)).
			WithArgs(contestID).
			WillReturnRows(statsRows)

		// Mock solutions query
		solutionsRows := sqlmock.NewRows([]string{
			"id",
			"participant_id",
			"participant_name",
			"state",
			"score",
			"penalty",
			"time_stat",
			"memory_stat",
			"language",
			"task_id",
			"task_position",
			"task_title",
			"contest_id",
			"contest_title",
			"updated_at",
			"created_at",
		}).AddRow(
			101,
			1,
			"Participant 1",
			5,
			100,
			0,
			1000,
			256,
			228,
			1,
			1,
			"Task 1",
			1,
			"Contest 1",
			now,
			now,
		)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(solutionsQuery)).
			WithArgs(contestID).
			WillReturnRows(solutionsRows)

		// Mock participants query
		participantsRows := sqlmock.NewRows([]string{
			"id",
			"name",
			"solved_in_total",
			"penalty_in_total",
		}).AddRow(1, "Participant 1", 2, 40)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(participantsQuery)).
			WithArgs(contestID, penalty, contestID).
			WillReturnRows(participantsRows)

		// Execute the function
		monitor, err := tf.contestRepo.ReadMonitor(context.Background(), contestID)
		require.NoError(t, err)
		require.Equal(t, expectedMonitor, monitor)
	})

	t.Run("contest not found", func(t *testing.T) {
		contestID := int32(999)

		// Mock statistics query to return no rows
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readStatisticsQuery)).
			WithArgs(contestID).
			WillReturnRows(sqlmock.NewRows([]string{"task_id", "position", "total", "success"}))

		// Mock solutions query
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(solutionsQuery)).
			WithArgs(contestID).
			WillReturnRows(sqlmock.NewRows([]string{
				"id", "participant_id", "participant_name", "state", "score",
				"penalty", "time_stat", "memory_stat", "language", "task_id",
				"task_position", "task_title", "contest_id", "contest_title",
				"updated_at", "created_at",
			}))

		// Mock participants query
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(participantsQuery)).
			WithArgs(contestID, int32(20), contestID).
			WillReturnRows(sqlmock.NewRows([]string{
				"id", "name", "solved_in_total", "penalty_in_total",
			}))

		// Execute the function
		monitor, err := tf.contestRepo.ReadMonitor(context.Background(), contestID)
		require.NoError(t, err)
		require.Empty(t, monitor.Summary)
		require.Empty(t, monitor.Participants)
	})

	t.Run("statistics query error", func(t *testing.T) {
		contestID := int32(1)
		dbError := errors.New("database connection failed")

		// Mock statistics query to return an error
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readStatisticsQuery)).
			WithArgs(contestID).
			WillReturnError(dbError)

		// Execute the function
		monitor, err := tf.contestRepo.ReadMonitor(context.Background(), contestID)
		require.Error(t, err)
		require.Contains(t, err.Error(), dbError.Error())
		require.Nil(t, monitor)
	})

	t.Run("solutions query error", func(t *testing.T) {
		contestID := int32(1)
		dbError := errors.New("solutions query failed")

		// Mock statistics query
		statsRows := sqlmock.NewRows([]string{"task_id", "position", "total", "success"}).
			AddRow(1, 1, 10, 5)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readStatisticsQuery)).
			WithArgs(contestID).
			WillReturnRows(statsRows)

		// Mock solutions query to return an error
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(solutionsQuery)).
			WithArgs(contestID).
			WillReturnError(dbError)

		// Execute the function
		monitor, err := tf.contestRepo.ReadMonitor(context.Background(), contestID)
		require.Error(t, err)
		require.Contains(t, err.Error(), dbError.Error())
		require.Nil(t, monitor)
	})

	t.Run("participants query error", func(t *testing.T) {
		contestID := int32(1)
		dbError := errors.New("participants query failed")

		// Mock statistics query
		statsRows := sqlmock.NewRows([]string{"task_id", "position", "total", "success"}).
			AddRow(1, 1, 10, 5)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readStatisticsQuery)).
			WithArgs(contestID).
			WillReturnRows(statsRows)

		// Mock solutions query
		solutionsRows := sqlmock.NewRows([]string{
			"id", "participant_id", "participant_name", "state", "score",
			"penalty", "time_stat", "memory_stat", "language", "task_id",
			"task_position", "task_title", "contest_id", "contest_title",
			"updated_at", "created_at",
		})

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(solutionsQuery)).
			WithArgs(contestID).
			WillReturnRows(solutionsRows)

		// Mock participants query to return an error
		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(participantsQuery)).
			WithArgs(contestID, int32(20), contestID).
			WillReturnError(dbError)

		// Execute the function
		monitor, err := tf.contestRepo.ReadMonitor(context.Background(), contestID)
		require.Error(t, err)
		require.Contains(t, err.Error(), dbError.Error())
		require.Nil(t, monitor)
	})
}

// Helper function to create int32 pointers
func i32p(i int32) *int32 {
	return &i
}

// Helper function to create string pointers
func sp(s string) *string {
	return &s
}

// Helper function to convert slice of interface{} to slice of driver.Value
func i2v(s []interface{}) []driver.Value {
	r := make([]driver.Value, len(s))
	for i, v := range s {
		r[i] = v
	}
	return r
}
