package repository

import (
	"context"
	"database/sql"
	"git.sch9.ru/new_gate/ms-tester/internal/models"
	"github.com/DATA-DOG/go-sqlmock"
	"github.com/jmoiron/sqlx"
	"github.com/stretchr/testify/require"
	"testing"
)

type problemTestFixture struct {
	db          *sql.DB
	sqlxDB      *sqlx.DB
	mock        sqlmock.Sqlmock
	problemRepo *ProblemRepository
}

func newProblemTestFixture(t *testing.T) *problemTestFixture {
	db, mock, err := sqlmock.New(sqlmock.QueryMatcherOption(sqlmock.QueryMatcherEqual))
	require.NoError(t, err)

	sqlxDB := sqlx.NewDb(db, "sqlmock")
	repo := NewProblemRepository(sqlxDB)

	return &problemTestFixture{
		db:          db,
		sqlxDB:      sqlxDB,
		mock:        mock,
		problemRepo: repo,
	}
}

// cleanup closes database connections
func (tf *problemTestFixture) cleanup() {
	tf.db.Close()
	tf.sqlxDB.Close()
}

func TestProblemRepository_CreateProblem(t *testing.T) {
	tf := newProblemTestFixture(t)
	defer tf.cleanup()

	t.Run("valid problem creation", func(t *testing.T) {
		title := "Test Problem"
		rows := sqlmock.NewRows([]string{"id"}).AddRow(1)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(createProblemQuery)).
			WithArgs(title).
			WillReturnRows(rows)

		id, err := tf.problemRepo.CreateProblem(context.Background(), tf.problemRepo.DB(), title)
		require.NoError(t, err)
		require.Equal(t, int32(1), id)
	})
}

func TestProblemRepository_ReadProblemById(t *testing.T) {
	tf := newProblemTestFixture(t)
	defer tf.cleanup()

	t.Run("valid problem read", func(t *testing.T) {
		id := int32(1)
		rows := sqlmock.NewRows([]string{"id", "title"}).
			AddRow(1, "Test Problem")

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(readProblemQuery)).
			WithArgs(id).
			WillReturnRows(rows)

		problem, err := tf.problemRepo.ReadProblemById(context.Background(), tf.problemRepo.DB(), id)
		require.NoError(t, err)
		require.NotNil(t, problem)
		require.Equal(t, int32(1), problem.Id)
		require.Equal(t, "Test Problem", problem.Title)
	})
}

func TestProblemRepository_DeleteProblem(t *testing.T) {
	tf := newProblemTestFixture(t)
	defer tf.cleanup()

	t.Run("valid problem deletion", func(t *testing.T) {
		id := int32(1)
		rows := sqlmock.NewResult(1, 1)

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(deleteProblemQuery)).
			WithArgs(id).
			WillReturnResult(rows)

		err := tf.problemRepo.DeleteProblem(context.Background(), tf.problemRepo.DB(), id)
		require.NoError(t, err)
	})
}

func TestProblemRepository_ListProblems(t *testing.T) {
	tf := newProblemTestFixture(t)
	defer tf.cleanup()

	t.Run("valid problems list", func(t *testing.T) {
		filter := models.ProblemsFilter{
			Page:     1,
			PageSize: 10,
		}
		listRows := sqlmock.NewRows([]string{"id", "title", "solved_count"}).
			AddRow(1, "Problem 1", 5).
			AddRow(2, "Problem 2", 3)
		countRows := sqlmock.NewRows([]string{"count"}).AddRow(2)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(ListProblemsQuery)).
			WithArgs(filter.PageSize, filter.Offset()).
			WillReturnRows(listRows)

		tf.mock.ExpectQuery(tf.sqlxDB.Rebind(CountProblemsQuery)).
			WillReturnRows(countRows)

		result, err := tf.problemRepo.ListProblems(context.Background(), tf.problemRepo.DB(), filter)
		require.NoError(t, err)
		require.NotNil(t, result)
		require.Len(t, result.Problems, 2)
		require.Equal(t, int32(1), result.Pagination.Page)
		require.Equal(t, int32(1), result.Pagination.Total)
	})
}

func TestProblemRepository_UpdateProblem(t *testing.T) {
	tf := newProblemTestFixture(t)
	defer tf.cleanup()

	t.Run("valid problem update", func(t *testing.T) {
		id := int32(1)
		problemUpdate := models.ProblemUpdate{
			Title:       sp("Updated Title"),
			TimeLimit:   i32p(1000),
			MemoryLimit: i32p(256),
		}
		rows := sqlmock.NewResult(1, 1)

		tf.mock.ExpectExec(tf.sqlxDB.Rebind(UpdateProblemQuery)).
			WithArgs(
				problemUpdate.Title,
				problemUpdate.TimeLimit,
				problemUpdate.MemoryLimit,
				nil, nil, nil, nil, nil,
				nil, nil, nil, nil, nil,
				id,
			).
			WillReturnResult(rows)

		err := tf.problemRepo.UpdateProblem(context.Background(), tf.problemRepo.DB(), id, problemUpdate)
		require.NoError(t, err)
	})
}
